/* RBLcheck.cmd */
/*
signal ON ERROR NAME verderhier;
signal ON FAILURE NAME verderhier;
signal ON HALT NAME verderhier;
signal ON NOVALUE NAME verderhier;
signal ON SYNTAX NAME verderhier;
signal ON NOTREADY NAME verderhier;
*/
Version="1.4"
parse arg Mail
parse version rxver1 rxver2 rxver3 rxver4 rxver5
parse source os invocation scriptfile
inifile= left(scriptfile, length(scriptfile)-3)'ini'
Log= 'c:\RBLcheck.log'
LogLevel= 2
ConLevel= 4
TriggerLevel= 1
RBLnames= ''
RBLs= ''
SkipIPs= ''
SkipLastHops= 0
HaltAtTrigger= 0
VersionStamp= 1

/* read ini file */
call linein inifile,1,0
do while (lines(inifile)> 0)
  iniline= space(translate(linein(inifile),' ','	'))
  parse var iniline iniword1 iniword2 iniword3 iniword4 iniword5
  iniwordt= translate(iniword1)
  select
    when iniwordt= 'LOGFILE' then
      Log= iniword2
    when iniwordt= 'LOGLEVEL' then
      LogLevel= iniword2
    when iniwordt= 'TRIGGERLEVEL' then
      TriggerLevel= iniword2
    when iniwordt= 'RBL' then
    do
      RBLs= RBLs' 'iniword2' 'iniword3' 'iniword4
      RBLnames= RBLnames' 'iniword2
    end
    when iniwordt= 'SKIPIP' then
      SkipIPs= SkipIPs' 'iniword2
    when iniwordt= 'SKIPLASTHOPS' then
      SkipLastHops= iniword2
    when iniwordt= 'HALTATTRIGGER' then
      HaltAtTrigger= iniword2
    when iniwordt= 'VERSIONSTAMP' then
      VersionStamp= iniword2
    when iniwordt= 'CONLEVEL' then
      ConLevel= iniword2
    otherwise
      nop
  end /* select */
end  
RBLs= space(RBLs)
RBLnames= space(RBLnames)
SkipIPs= space(SkipIPs)

call WriteLog 1,'--- RBLcheck started, version 'Version', loglevel 'LogLevel
call WriteLog 4,'   OS: 'os
call WriteLog 4,'   Invocation: 'invocation
call WriteLog 4,'   Script: 'scriptfile
call WriteLog 4,'   REXX: 'rxver1 rxver2 rxver3 rxver4 rxver5
call WriteLog 4,'   Arg: 'Mail
call WriteLog 4,'   Ini: 'inifile
call WriteLog 3,'  TriggerLevel: 'TriggerLevel
call WriteLog 3,'  RBLs: 'RBLnames
if words(SkipIPs)>1 then
  call WriteLog 3,'  SkipIP: 'SkipIPs
else
  call WriteLog 3,'  SkipIP: none'
call WriteLog 3,'  SkipLastHops: 'SkipLastHops
call WriteLog 3,'  HaltAtTrigger: 'HaltAtTrigger
call WriteLog 3,'  VersionStamp: 'VersionStamp
if words(Mail)<1 then
do
  call WriteLog 1,"No Arg: don't know what mail to check... exiting."
  call WriteLog 1,'--- RBLcheck finished'
  exit
end

call RxFuncAdd "SockLoadFuncs","RxSock","SockLoadFuncs"
call SockLoadFuncs

/* read mail headers into memory, extract IP addresses */
IPaddresses= ''
maiq= rxqueue('Create')
oldq= rxqueue('set',maiq)
head= 0
FullLine= ''
do while (lines(Mail)> 0) & (head= 0)
  line= linein(Mail)
  if words(line)= 0 then
  do
    head= 1
    IPaddresses= IPaddresses' 'InterpretLine(FullLine)
  end
  else
  do
    firstchar= left(line,1)
    if verify(firstchar,' 	','M')= 0 then
    do
      if \(pos('X-RBLCHECK',translate(line))= 1) then
        queue line
      /* previous (folded) header now complete... check it */
      FullLine= space(translate(FullLine,' ','	'))
      if words(FullLine)> 0 then
        IPaddresses= IPaddresses' 'InterpretLine(FullLine)
      FullLine= line
    end
    else
    do
      /* unfolding header */
      FullLine= FullLine' 'line
      queue line
    end
  end
end
IPaddresses= space(IPaddresses)

if head= 0 then
do
  call WriteLog 1,"No blank line between headers and message body (RFC822)... exiting."
  call WriteLog 1,'--- RBLcheck finished'
  exit
end

if VersionStamp= 1 then
  queue 'X-RBLcheck-version: 'Version' (see http://os2.besse.nl/)'

/* check IP addresses */
rblscore= 0
rblhit= 0
do while words(IPaddresses)> 0
  RBLrun= RBLs
  parse var IPaddresses line IPaddresses
  do while (words(RBLrun)> 0)
    parse var RBLrun iniword2 iniword3 iniword4 RBLrun
    if CheckRBL(line, iniword2, iniword3, iniword4)> 0 then
    do
      rblscore= rblscore+ iniword4
      rblhit= 1
      if HaltAtTrigger= 1 then
        if rblscore >= TriggerLevel then
        do
          IPaddresses= ""
          RBLrun= ""
        end
    end
  end  
end

/* add RBLcheck headers to headers in memory */
if rblhit= 0 then
  do
    queue 'X-RBLcheck: negative'
    call WriteLog 1,'RBLcheck negative: no hits.'
  end
else
do
  queue 'X-RBLcheck: positive'
  call WriteLog 1,'RBLcheck positive: some hits.'
  if HaltAtTrigger= 0 then
  do
    queue 'X-RBLcheck-score: 'rblscore
    call WriteLog 2,' Final score: 'rblscore
  end
  if rblscore < TriggerLevel then
    queue 'X-RBLcheck-trigger-reached: false'
  else
  do
    queue 'X-RBLcheck-trigger-reached: true'
    call WriteLog 1,'Trigger reached.'
  end
end
queue

/* read message body into memory */
do while lines(Mail)> 0
  queue linein(Mail)
end

/* write back mail message */
call lineout Mail,,1
do while queued()> 0
  parse pull line
  call lineout Mail,line
end
call lineout Mail

call rxqueue 'Delete',maiq
call rxqueue 'Set',oldq

call WriteLog 1,'--- RBLcheck finished'

exit

/* subroutines */
WriteLog:
if LogLevel>= arg(1) then
  call lineout Log,date('S')' 'time()' 'arg(2)
if ConLevel>= arg(1) then
  say date('S')' 'time()' 'arg(2)
return

InterpretLine:
mline= arg(1)
hitaddress= ''
uline= translate(mline)
firstword= word(uline,1)
select
  when firstword= 'SUBJECT:' then
    call WriteLog 1,mline
  when firstword= 'FROM:' then
    call WriteLog 2,' 'mline
  when firstword= 'TO:' then
    call WriteLog 3,'  'mline
  when firstword= 'RECEIVED:' then
  do
    /* check only the from part */
    parse var uline dummy ' FROM ' line1
    parse var line1 line2 ' BY ' dummy
    parse var line2 line3 ' VIA ' dummy
    parse var line3 line4 ' WITH ' dummy
    parse var line4 line5 ' ID ' dummy
    parse var line5 line6 ' FOR ' dummy
    parse var line6 linef ';' dummy
    linef= translate(linef, ' ', '[');
    linef= translate(linef, ' ', '(');
    linef= translate(linef, ' ', ']');
    linef= translate(linef, ' ', ')');
    /* find IP address */
    IPaddress= 0
    do while words(linef)> 0
      parse var linef woord linef
      if verify(woord,'1234567890.')= 0 then
      do
        parse var woord byte1 '.' byte2 '.' byte3 '.' byte4
        if (datatype(byte1,'W'))&(datatype(byte2,'W'))&(datatype(byte3,'W'))&(datatype(byte4,'W')) then
          IPaddress= space(woord)
      end
    end
    if \(IPaddress= 0) then
    do
      if SkipLastHops> 0 then
      do
        call WriteLog 3,'  Received from: 'IPaddress', skipping this hop.'
        SkipLastHops= SkipLastHops- 1
      end
      else
      do
        select
          when byte1< 1 then
            call WriteLog 4,'   Received from: 'IPaddress', skipping, class A reserved range.'
          when byte1= 10 then
            call WriteLog 4,'   Received from: 'IPaddress', skipping, private class A range.'
          when byte1= 127 then
            call WriteLog 4,'   Received from: 'IPaddress', skipping, loopback range.'
          when (byte1= 172) & (byte2> 15) & (byte2< 32) then
            call WriteLog 4,'   Received from: 'IPaddress', skipping, private class B range.'
          when (byte1= 192) & (byte2= 168) then
            call WriteLog 4,'   Received from: 'IPaddress', skipping, private class C range.'
          when (byte1> 223) & (byte1< 240) then
            call WriteLog 4,'   Received from: 'IPaddress', skipping, class D multicast range.'
          when byte1> 239 then
            call WriteLog 4,'   Received from: 'IPaddress', skipping, class E reserved range.'
          when pos(IPaddress, SkipIPs)> 0 then
            call WriteLog 3,'  Received from: 'IPaddress', skipping as requested in .ini.'
          when pos(IPaddress, IPaddresses)> 0 then
            call WriteLog 4,'   Received from: 'IPaddress', skipping, already queued for testing.'
          otherwise
          do
            hitaddress= IPaddress
            call WriteLog 3,'  Received from: 'IPaddress', queued for testing.'
          end
        end /* select */
      end
    end
  end
  otherwise
    nop
end /* select */
return hitaddress

CheckRBL:
IP= arg(1)
RBLname= arg(2)
RBLdns= arg(3)
weight= arg(4)
CheckRBLhit= 0
/* reverse bytes of IP address */
parse var IP byte1 '.' byte2 '.' byte3 '.' byte4
rIP= byte4'.'byte3'.'byte2'.'byte1
GHBNhost.!addr= '-'
/* check at RBL */
call Writelog 4,'   call GetHostByName ('rIP'.'RBLdns').'
rc=SockGetHostByName(rIP'.'RBLdns, "GHBNhost.!")
call Writelog 4,'   GetHostByName ('rIP'.'RBLdns'): 'GHBNhost.!addr', errno: 'errno', h_errno: 'h_errno', rc: 'rc'.'
if errno= 0 then
do
  if rc= 1 then
  do
    CheckRBLhit= 1
    queue 'X-RBLcheck-hits: 'IP' found at 'RBLname' (score: 'weight')'
    call Writelog 2,' Hit: 'IP' found at 'RBLname': 'GHBNhost.!addr' (score: 'weight')'
  end
end
else
do
  call Writelog 1,'Error while checking RBL 'RBLname': 'errno
end
drop GHBNhost.
return CheckRBLhit

verderhier:
call Writelog 1,'--- Error on line 'SIGL'! Aborting.'
exit
